/* SPDX-License-Identifier: LGPL-2.1-or-later */
/*
 * Copyright (C) 2024 Red Hat, Inc.
 */

#include "libnm-core-impl/nm-default-libnm-core.h"

#include "nm-initrd-generator.h"

#include <libnvme.h>

#include "libnm-log-core/nm-logging.h"
#include "libnm-core-intern/nm-core-internal.h"

/*****************************************************************************/

#define _NMLOG(level, domain, ...) \
    nm_log((level),                \
           (domain),               \
           NULL,                   \
           NULL,                   \
           "nbft-reader: " _NM_UTILS_MACRO_FIRST(__VA_ARGS__) _NM_UTILS_MACRO_REST(__VA_ARGS__))

/*****************************************************************************/

#define VALID_ADDR(s) \
    (strlen(s) > 0 && g_strcmp0(s, "0.0.0.0") != 0 && g_strcmp0(s, "::") != 0)


static NMConnection *
parse_hfi(struct nbft_info_hfi *hfi, int iface_idx, char **hostname) {
    NMConnection                         *connection;
    NMSetting                            *s_wired      = NULL;
    gs_free char                         *hwaddr       = NULL;
    gs_free char                         *ifname       = NULL;
    NMSetting                            *s_ip4        = NULL;
    NMSetting                            *s_ip6        = NULL;
    nm_auto_unref_ip_address NMIPAddress *ipaddr       = NULL;
    nm_auto_unref_ip_address NMIPAddress *gateway      = NULL;
    gs_free_error GError                 *error        = NULL;
    int                                   family       = AF_UNSPEC;

    /* Pre-checks */
    hwaddr = g_strdup_printf("%02X:%02X:%02X:%02X:%02X:%02X",
                             hfi->tcp_info.mac_addr[0],
                             hfi->tcp_info.mac_addr[1],
                             hfi->tcp_info.mac_addr[2],
                             hfi->tcp_info.mac_addr[3],
                             hfi->tcp_info.mac_addr[4],
                             hfi->tcp_info.mac_addr[5]);

    if (!nm_utils_hwaddr_valid(hwaddr, -1)) {
        _LOGW(LOGD_CORE, "NBFT: invalid MAC address '%s'.", hwaddr);
        return NULL;
    }

    /* The libnvme public NBFT API only provides decoded string,
     * need to make an educated guess.
     */
    if (strchr(hfi->tcp_info.ipaddr, ':'))
        family = AF_INET6;
    else
    if (strchr(hfi->tcp_info.ipaddr, '.'))
        family = AF_INET;
    else {
        _LOGW(LOGD_CORE, "NBFT: unable to detect address family for '%s'.", hfi->tcp_info.ipaddr);
        return NULL;
    }

    if (!nm_utils_ipaddr_valid(family, hfi->tcp_info.ipaddr)) {
        _LOGW(LOGD_CORE, "NBFT: invalid IP address '%s'.", hfi->tcp_info.ipaddr);
        return NULL;
    }

    /* Create new connection */
    connection = nm_simple_connection_new();

    ifname = g_strdup_printf("nbft%d", iface_idx);
    nm_connection_add_setting(connection,
                              g_object_new(NM_TYPE_SETTING_CONNECTION,
                                           NM_SETTING_CONNECTION_TYPE,
                                           NM_SETTING_WIRED_SETTING_NAME,
                                           NM_SETTING_CONNECTION_ID,
                                           ifname,
                                           NM_SETTING_CONNECTION_INTERFACE_NAME,
                                           ifname,
                                           NM_SETTING_CONNECTION_AUTOCONNECT_PRIORITY,
                                           NMI_AUTOCONNECT_PRIORITY_FIRMWARE,
                                           NULL));

    /* MAC address */
    s_wired = nm_setting_wired_new();
    g_object_set(s_wired,
                 NM_SETTING_WIRED_MAC_ADDRESS,
                 hwaddr,
                 NULL);

    nm_connection_add_setting(connection, s_wired);

    /* VLAN ID */
    if (hfi->tcp_info.vlan > 0) {
        NMSetting *s_vlan = nm_setting_vlan_new();

        g_object_set(s_vlan,
                     NM_SETTING_VLAN_ID,
                     hfi->tcp_info.vlan,
                     NULL);

        nm_connection_add_setting(connection, s_vlan);
    }

    /* IP addresses */
    s_ip4 = nm_setting_ip4_config_new();
    s_ip6 = nm_setting_ip6_config_new();

    switch (family) {
        /* IPv4 */
        case AF_INET:
            g_object_set(s_ip6,
                         NM_SETTING_IP_CONFIG_METHOD,
                         NM_SETTING_IP6_CONFIG_METHOD_DISABLED,
                         NULL);
            if (VALID_ADDR(hfi->tcp_info.dhcp_server_ipaddr)) {
                g_object_set(s_ip4,
                             NM_SETTING_IP_CONFIG_METHOD,
                             NM_SETTING_IP4_CONFIG_METHOD_AUTO,
                             NULL);
                if (hfi->tcp_info.host_name && strlen(hfi->tcp_info.host_name) > 0)
                    g_object_set(s_ip4,
                                 NM_SETTING_IP_CONFIG_DHCP_HOSTNAME,
                                 hfi->tcp_info.host_name,
                                 NULL);
            } else {
                g_object_set(s_ip4,
                             NM_SETTING_IP_CONFIG_METHOD,
                             NM_SETTING_IP4_CONFIG_METHOD_MANUAL,
                             NULL);
                ipaddr = nm_ip_address_new(AF_INET, hfi->tcp_info.ipaddr, hfi->tcp_info.subnet_mask_prefix, &error);
                if (!ipaddr) {
                    _LOGW(LOGD_CORE, "Cannot parse IP address string '%s': %s", hfi->tcp_info.ipaddr, error->message);
                    return NULL;
                }
                nm_setting_ip_config_add_address(NM_SETTING_IP_CONFIG(s_ip4), ipaddr);
                if (VALID_ADDR(hfi->tcp_info.gateway_ipaddr))
                    g_object_set(s_ip4,
                                 NM_SETTING_IP_CONFIG_GATEWAY,
                                 hfi->tcp_info.gateway_ipaddr,
                                 NULL);
                if (hfi->tcp_info.route_metric > 0)
                    g_object_set(s_ip4,
                                 NM_SETTING_IP_CONFIG_ROUTE_METRIC,
                                 hfi->tcp_info.route_metric,
                                 NULL);
                if (VALID_ADDR(hfi->tcp_info.primary_dns_ipaddr))
                    nm_setting_ip_config_add_dns(NM_SETTING_IP_CONFIG(s_ip4),
                                                 hfi->tcp_info.primary_dns_ipaddr);
                if (VALID_ADDR(hfi->tcp_info.secondary_dns_ipaddr))
                    nm_setting_ip_config_add_dns(NM_SETTING_IP_CONFIG(s_ip4),
                                                 hfi->tcp_info.secondary_dns_ipaddr);
            }
            break;

        /* IPv6 */
        case AF_INET6:
            g_object_set(s_ip4,
                         NM_SETTING_IP_CONFIG_METHOD,
                         NM_SETTING_IP4_CONFIG_METHOD_DISABLED,
                         NULL);
            if (VALID_ADDR(hfi->tcp_info.dhcp_server_ipaddr)) {
                g_object_set(s_ip6,
                             NM_SETTING_IP_CONFIG_METHOD,
                             NM_SETTING_IP6_CONFIG_METHOD_AUTO,
                             NULL);
                if (hfi->tcp_info.host_name && strlen(hfi->tcp_info.host_name) > 0)
                    g_object_set(s_ip6,
                                 NM_SETTING_IP_CONFIG_DHCP_HOSTNAME,
                                 hfi->tcp_info.host_name,
                                 NULL);
            } else {
                g_object_set(s_ip6,
                             NM_SETTING_IP_CONFIG_METHOD,
                             NM_SETTING_IP6_CONFIG_METHOD_MANUAL,
                             NULL);
                ipaddr = nm_ip_address_new(AF_INET6,
                                           hfi->tcp_info.ipaddr,
                                           hfi->tcp_info.subnet_mask_prefix > 0 ? hfi->tcp_info.subnet_mask_prefix : 128,
                                           &error);
                if (!ipaddr) {
                    _LOGW(LOGD_CORE, "Cannot parse IP address string '%s': %s", hfi->tcp_info.ipaddr, error->message);
                    return NULL;
                }
                nm_setting_ip_config_add_address(NM_SETTING_IP_CONFIG(s_ip6), ipaddr);
                if (VALID_ADDR(hfi->tcp_info.gateway_ipaddr))
                    g_object_set(s_ip6,
                                 NM_SETTING_IP_CONFIG_GATEWAY,
                                 hfi->tcp_info.gateway_ipaddr,
                                 NULL);
                if (hfi->tcp_info.route_metric > 0)
                    g_object_set(s_ip6,
                                 NM_SETTING_IP_CONFIG_ROUTE_METRIC,
                                 hfi->tcp_info.route_metric,
                                 NULL);
                if (VALID_ADDR(hfi->tcp_info.primary_dns_ipaddr))
                    nm_setting_ip_config_add_dns(NM_SETTING_IP_CONFIG(s_ip6),
                                                 hfi->tcp_info.primary_dns_ipaddr);
                if (VALID_ADDR(hfi->tcp_info.secondary_dns_ipaddr))
                    nm_setting_ip_config_add_dns(NM_SETTING_IP_CONFIG(s_ip6),
                                                 hfi->tcp_info.secondary_dns_ipaddr);
            }
            break;
        default:
            g_warn_if_reached();
    }

    nm_connection_add_setting(connection, s_ip4);
    nm_connection_add_setting(connection, s_ip6);

    /* Hostname */
    if (hfi->tcp_info.host_name && strlen(hfi->tcp_info.host_name) > 0) {
        free(*hostname);
        *hostname = strdup(hfi->tcp_info.host_name);
    }

    /* TODO: translate the following HFI struct members?
     *         hfi->tcp_info.pci_sbdf
     *         hfi->tcp_info.ip_origin
     *         hfi->tcp_info.dhcp_server_ipaddr
     *         hfi->tcp_info.this_hfi_is_default_route
     *         hfi->tcp_info.dhcp_override
     */

    if (!nm_connection_normalize(connection, NULL, NULL, &error)) {
        _LOGW(LOGD_CORE, "Generated an invalid connection: %s", error->message);
        nm_clear_pointer(&connection, g_object_unref);
    }
    return connection;
}


NMConnection **
nmi_nbft_reader_parse(const char *sysfs_dir, char **hostname)
{
    GPtrArray            *a;
    gs_free char         *path  = NULL;
    gs_free_error GError *error = NULL;
    GDir                 *dir;
    const char           *entry_name;
    int                   idx   = 0;

    g_return_val_if_fail(sysfs_dir != NULL, NULL);
    path = g_build_filename(sysfs_dir, "firmware", "acpi", "tables", NULL);

    dir = g_dir_open(path, 0, &error);
    if (!dir) {
        _LOGW(LOGD_CORE, "Can't open %s: %s", path, error->message);
        return NULL;
    }

    a = g_ptr_array_new();

    while ((entry_name = g_dir_read_name(dir))) {
        gs_free char *entry_path = NULL;
        struct nbft_info *nbft;
        struct nbft_info_hfi **hfi;
        int ret;

        if (!g_str_has_prefix(entry_name, "NBFT"))
            continue;

        entry_path = g_build_filename(path, entry_name, NULL);
        ret = nvme_nbft_read(&nbft, entry_path);
        if (ret) {
            _LOGW(LOGD_CORE, "Error parsing NBFT table %s: %m", entry_path);
            continue;
        }

        for (hfi = nbft->hfi_list; hfi && *hfi; hfi++, idx++) {
            NMConnection *connection;

            if (g_strcmp0 ((*hfi)->transport, "tcp") != 0) {
                _LOGW(LOGD_CORE, "NBFT table %s, HFI descriptor %d: unsupported transport type '%s'",
                      entry_path, (*hfi)->index, (*hfi)->transport);
                continue;
            }

            connection = parse_hfi(*hfi, idx, hostname);
            if (connection)
                g_ptr_array_add(a, connection);
        }

        nvme_nbft_free(nbft);
    }

    g_dir_close(dir);
    g_ptr_array_add(a, NULL);  /* trailing NULL-delimiter */
    return (NMConnection **) g_ptr_array_free(a, FALSE);
}
